"""A list of reusable constants for the package.

MEDIA_FIELDS (tuple): Fields that can be set to file names of allowable
    media types for the given field.
LANGUAGE_DEPENDENT_FIELDS (tuple): Fields for which values can vary by
    language. A single ODK XLSForm can have many such fields, suffixed
    by '::language'.
TRUNCATABLE_FIELDS (tuple): Fields that should be limited to a specific
    length. Current limit is 100 chars, which is somewhat arbitrary but
    turns out good in converted forms.
"""
SYNTAX = {
    'xlsforms': {
        'language_field_delimiters': [':', '::']
    }
}
IGNORE_RELEVANT_TOKEN = '#####'
SUPPORTED_FORMATS = ('html', 'doc')
LANGUAGE_PERTINENT_WORKSHEETS = ('survey', 'choices', 'external_choices')
XLSFORM_SUPPORTED_MULTIMEDIA_TYPES = ['image', 'audio', 'video']
# MEDIA_FIELDS = (image, media:image, media::image, ...)
MEDIA_FIELDS = tuple(x for x in XLSFORM_SUPPORTED_MULTIMEDIA_TYPES) + \
   tuple('media' + y + x for x in XLSFORM_SUPPORTED_MULTIMEDIA_TYPES
         for y in SYNTAX['xlsforms']['language_field_delimiters'])
# for x in XLSFORM_SUPPORTED_MULTIMEDIA_TYPES:
#     MEDIA_FIELDS.append(x)
#     for y in SYNTAX['xlsforms']['language_field_delimiters']:
#         MEDIA_FIELDS.append('media' + y + x)
CONSTRAINT_MESSAGE_VARIATIONS = \
    tuple('constraint{}message'.format(x) for x in [' ', '_'])
LANGUAGE_DEPENDENT_FIELDS_NONMEDIA_FIELDS = ('label', 'hint', 'ppp_input') + \
                            CONSTRAINT_MESSAGE_VARIATIONS
LANGUAGE_DEPENDENT_FIELDS = LANGUAGE_DEPENDENT_FIELDS_NONMEDIA_FIELDS + \
                            MEDIA_FIELDS
RELEVANCE_FIELD_TOKENS = ('relevant', 'relevance')
TRUNCATABLE_FIELDS = ('constraint',) + RELEVANCE_FIELD_TOKENS
MULTI_ARGUMENT_CONVERSION_OPTIONS = ('preset', 'format', 'language')
PPP_REPLACEMENTS_FIELDS = ('label',) + RELEVANCE_FIELD_TOKENS
CHOICE_NAME_VARIATIONS = ('name', 'value')
PRESETS = {
    'full': {
        'field_replacements': (),
        'field_exclusions': (),
        'other_specific_exclusions': (),
        'general_exclusions': False,
        'render_settings': {
            'html': {
                'side_letters': True
            }
        }
    },
    'internal': {
        'field_replacements': (),
        'field_exclusions': (),
        'other_specific_exclusions': (),
        'general_exclusions': False,
        'render_settings': {
            'html': {
                'side_letters': True
            }
        }
    },
    'minimal': {
        'field_replacements': PPP_REPLACEMENTS_FIELDS,
        'field_exclusions':
        ('constraint', 'constraint_message', 'type') + CHOICE_NAME_VARIATIONS,
        'other_specific_exclusions': ('choice names',),
        'general_exclusions': True,
        'render_settings': {
            'html': {
                'side_letters': False
            }
        }
    },
    'public': {
        'field_replacements': (),
        'field_exclusions': (),
        'other_specific_exclusions': (),
        'general_exclusions': False,
        'render_settings': {
            'html': {
                'side_letters': True
            }
        }
    },
}
PRESETS['developer'] = PRESETS['full']  # alias
PRESETS['standard'] = PRESETS['minimal']  # alias
PRESETS['basic'] = PRESETS['minimal']  # alias



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Command Line Interface."""
from argparse import ArgumentParser
from sys import stderr
from copy import copy

from ppp import run
from ppp.definitions.constants import SUPPORTED_FORMATS
from ppp.definitions.abstractions import chain
from ppp.definitions.error import OdkException, OdkFormError


def _required_fields(parser):
    """Add required fields to parser.

    Args:
        parser (ArgumentParser): Argparse object.

    Returns:
        ArgumentParser: Argeparse object.
    """
    # File
    #   type:'string'
    file_help = 'Path to source XLSForm(s).'
    parser.add_argument('xlsxfiles', nargs='+', help=file_help)
    # Bundled Option Presets
    #   type='single selection', options:'custom, developer/full, internal,
    #   public', default:'public'
    #   note='the CLI does not have 'custom'. Leaving this option blank in the
    #   CLI os the same as 'custom'.'
    presets_help = \
        ('Select from a preset of bundled options. The \'developer\' or '
         '\'full\' preset'
         ' renders a form that is the most similar to the original XlsForm. '
         'The \'internal\' preset is more human readable but is not stripped '
         'of sensitive information. The \'public\' option is like the '
         '\'internal\' optoin, only with sensitive information removed.')
    parser.add_argument('-p', '--preset', nargs='+',
                        choices=('public', 'internal', 'full', 'developer',
                                 'minimal'),
                        default='full', help=presets_help)
    return parser


def _non_preset_optional_fields(parser):
    """Add non-preset optional fields to parser.

    Args:
        parser (ArgumentParser): Argparse object.

    Returns:
        ArgumentParser: Argeparse object.
    """
    # Language
    #   type='single selection', default:''
    language_help = \
        ('Language to write the paper version in. If not specified, the '
         '\'default_language\' in the \'settings\' worksheet is used. If that '
         'is not specified and more than one language is in the XLSForm, the '
         'language that comes first alphabetically will be used.')
    parser.add_argument('-l', '--language', nargs='+', help=language_help)

    # Output Format
    #   type='single selection', default:'html'
    format_help = ('File format. HTML and DOC are supported formats. PDF is '
                   'not supported, but one can easily convert a PPP .doc file '
                   'into PDF via the use of WKHTMLtoPDF '
                   '(https://wkhtmltopdf.org/).')
    parser.add_argument('-f', '--format', nargs='+',
                        choices=SUPPORTED_FORMATS, default='doc',
                        help=format_help)
    # Choose Template
    #   type='single selection', default:'default'
    template_help = ('Choose template to render results.')
    parser.add_argument('-t', '--template', nargs = '+',
                        choices = ('default', 'old', 'test'),
                        default = 'default',
                        help = template_help)
    return parser


def _preset_optional_fields(parser):
    """Add preset options to parser.

    These options are all boolean toggles. Their value should automatically
    update based on the bundle option preset selected. Also, they should either
    (a) greyed out unless the 'custom bundle option preset is selected, or (b)
    if a bundle option preset other than 'custom' is selected, the bundle
    option preset should automatically change to 'custom' if any of the
    options are changed.


    Args:
        parser (ArgumentParser): Argparse object.

    Returns:
        ArgumentParser: Argeparse object.
    """
    # TODO: Input replacement
    #   type='boolean', default:'TRUE if public else FALSE if internal else
    #   FALSE if developer/full'
    # input_replacement_help = \
    #     ('Adding this option will toggle replacement of visible choice '
    #      'options in input fields. Instead of the normal choice options, '
    #      'whatever has been placed in the \'ppp_input\' field of the XlsForm '
    #      'will be used. This is normally to hide sensitive information.')
    # parser.add_argument('-i', '--input-replacement', action='store_true',
    #                     help=input_replacement_help)

    # TODO: Exclusion
    #   type='boolean', default:'TRUE if public else FALSE if internal else
    #   FALSE if developer/full
    # exclusion_help = \
    #     ('Adding this option will toggle exclusion of certain survey form '
    #      'compoments from the rendered form. This can be used to remove '
    #      'ODK-specific implementation elements from the form which are only '
    #      'useful for developer, and can also be used to wholly remove '
    #      'sensitive information without any replacement.')
    # parser.add_argument('-e', '--exclusion', action='store_true',
    #                     help=exclusion_help)

    # TODO: Human-readable relevant text
    #   type='boolean', default:'TRUE if public else TRUE if internal else
    #   FALSE if developer/full
    # hr_relevant_help = \
    #     ('Adding this option will toggle display of human readable '
    #      '\'relevant\' text, rather than the syntax-heavy codified logic of '
    #      'the original XlsForm.')
    # parser.add_argument('-r', '--hr-relevant', action='store_true',
    #                     help=hr_relevant_help)

    # TODO: Human-readable constraint text
    #   type='boolean', default:'TRUE if public else FALSE if internal else
    #   FALSE if developer/full
    # hr_constraint_help = \
    #     ('Adding this option will toggle display of human readable '
    #      '\'constraint\' text, rather than the syntax-heavy codified logic of '
    #      'the original XlsForm.')
    # parser.add_argument('-c', '--hr-constraint', action='store_true',
    #                     help=hr_constraint_help)

    # TODO: No constraint text
    #   type='boolean', default:'FALSE if public else FALSE if internal
    #   else FALSE if developer/full
    # no_constraint_help = \
    #     ('Adding this option will toggle removal of all constraints from the '
    #      'rendered form.')
    # parser.add_argument('-C', '--no-constraint', action='store_true',
    #                     help=no_constraint_help)

    # TODO: General text replacements
    #   type='boolean', default:'TRUE if public else TRUE if internal else
    #   FALSE if developer/full
    # text_replacements_help = \
    #     ('Adding this option will toggle text replacements as shown in the '
    #      '\'text_replacements\' worksheet of the XlsForm. The most common '
    #      'function of text replacement is to render more human readable '
    #      'variable names, but can also be used to remove sensitive information'
    #      'or add brevity or clarity where needed.')
    # parser.add_argument('-T', '--text-replacements', action='store_true',
    #                     help=text_replacements_help)
    return parser


def _cli_only_fields(parser):
    """Add CLI-only fields to parser.

    Args:
        parser (ArgumentParser): Argparse object.

    Returns:
        ArgumentParser: Argeparse object.
    """
    debug_help = \
        'Turns on debug mode. Currently only works for \'html\' format. Only' \
        ' feature of debug mode currently is that it prints a stringified ' \
        'JSON representation of survey to the JavaScript console.'
    parser.add_argument('-d', '--debug', action='store_true', help=debug_help)

    # Survey Form Component Highlighting
    highlighting_help = 'Turns on highlighting of various portions of survey' \
                        ' components. Useful to assess positioning.'
    parser.add_argument('-H', '--highlight', action='store_true',
                        help=highlighting_help)

    # Out path
    out_help = ('Path (including file name) to save converted file if 1 file, '
                'else path to directory for multiple files, in which case file'
                ' names will be automatically generated.\n\nIf this argument '
                'is not supplied, then STDOUT is used.')
    parser.add_argument('-o', '--outpath', help=out_help)
    return parser


def _add_arguments(parser):
    """Add arguments to parser.

    Args:
        parser (ArgumentParser): Argparse object.

    Returns:
        ArgumentParser: Argeparse object.
    """
    return chain(parser, funcs=[_cli_only_fields, _non_preset_optional_fields,
                                _preset_optional_fields, _required_fields])


def cli():
    """Command line interface for package.

    Side Effects: Executes program.

    Command Syntax: python3 -m ppp <file> <options>

    Examples:
        # Creates a 'myFile.html' in English with component highlighting.
        python3 -m ppp myFile.xlsx -l 'English' -h > myFile.html
    """
    prog_desc = 'Convert XLSForm to Paper version.'

    argeparser = ArgumentParser(description=prog_desc)
    parser = _add_arguments(copy(argeparser))
    args = parser.parse_args()

    if args.highlight and args.format and args.format not in SUPPORTED_FORMATS:
        msg = 'Can only specify highlighting when using the following ' \
              'formats: \'html\', \'pdf\'.'
        raise OdkFormError(msg)

    try:
        run(files=list(args.xlsxfiles),
            languages=[l for l in args.language] if args.language else [None],
            format=args.format, outpath=args.outpath,
            debug=args.debug, highlight=args.highlight, preset=args.preset, template=args.template)
    except OdkException as err:
        err = 'An error occurred while attempting to convert \'{}\':\n{}'\
            .format(args.xlsxfiles, err)
        print(err, file=stderr)


if __name__ == '__main__':
    cli()



"""Utils."""
from ..config import EXCLUSION_TOKEN
from .constants import PRESETS


def exclusion(item, settings):
    """Identify item as to be excluded or not.

    Identify whether item should be excluded from rendering based on whether
    there is presence of exclusion token in the ppp_excludes field of an
    XlsForm, assuming the user has determined that there should be exclusions,
    based on explicit 'exclusion' parameter, or via certain preset options.

    Args:
        item (object): Object of type OdkPrompt, OdkGroup, OdkRepeat, or
            OdkTable.
        settings (dict): Keyword argument settings passed by user to PPP for
            conversion.

    Returns
        bool: True if item should be excluded, else False.
    """
    config = {
        'error_on_no_exclude_column_for_preset': False
    }
    exclude = False

    try:
        if 'exclusion' in settings or 'preset' in settings \
                and PRESETS[settings['preset']]['general_exclusions']:

            if hasattr(item, 'row'):
                item_data = 'row'
            elif hasattr(item, 'opener'):
                item_data = 'opener'
            else:
                # Table; Rather than explicitly exluding a table, the group
                # itself should be excluded.
                return False

            token = EXCLUSION_TOKEN.lower()
            exclude = getattr(item, item_data)['ppp_excludes'].lower() == token

        return exclude
    except KeyError:
        err = True
        if config['error_on_no_exclude_column_for_preset'] == False \
                and 'preset' in settings:
            err = False
        if err:
            msg = 'If using exclusion option or preset with exclusions, ' \
                  'XlsForm must have field named \'ppp_excludes\'.'
            raise KeyError(msg)
        else:
            return exclude



"""Module for the OdkForm class."""
import os

#from ppp.config import TEMPLATE_ENV
from ppp.config import get_template_env
from ppp.definitions.error import OdkFormError
from ppp.odkcalculate import OdkCalculate
from ppp.odkchoices import OdkChoices
from ppp.odkgroup import OdkGroup, set_template_env as odkgroup_template
from ppp.odkprompt import OdkPrompt, set_template_env as odkpromt_template
from ppp.odkrepeat import OdkRepeat, set_template_env as odkrepeat_template
from ppp.odktable import OdkTable, set_template_env as odktable_template
from ppp.definitions.utils import exclusion
from pmix.xlsform import Xlsform

TEMPLATE_ENV = None

def set_template_env(template):
  global TEMPLATE_ENV
  TEMPLATE_ENV = get_template_env(template)
  odkgroup_template(template)
  odkpromt_template(template)
  odkrepeat_template(template)
  odktable_template(template)

class OdkForm:
    """Class to represent an entire XLSForm.

    Attributes:
        settings (dict): A dictionary representation of the original 'settings'
            worksheet of an ODK XLSForm.
        title (str): Title of the ODK form.
        choices (dict): A list of rows from the 'choices' worksheet.
        ext_choices (dict): A list of rows from the 'external_choices'
            worksheet.
        metadata (dict): A dictionary of metadata for the original and
            converted ODK forms.
        questionnaire (list): An ordered representation of the ODK form,
            comprised of OdkPrompt, OdkGroup, OdkRepeat, and OdkTable objects.

    """

    def __init__(self, wb):
        """Initialize the OdkForm.

        Create an instance of an ODK form, including survey representation,
        choice options, settings, and metadata.

        Args:
            wb (Xlsform): A Xlsform object meeting XLSForm specification.

        Raises:
            OdkformError: No ODK form is supplied.
        """
        self.settings = wb.settings
        self.language = wb.form_language

        self.title = self.get_title(settings=self.settings, wb=wb)
        self.metadata = {  # TODO Finish filling this out.
            'last_author': str(),
            'last_updated': str(),
            'changelog': None,
            'info': None,
            'raw_data': wb
        }
        self.choices = self.get_choices(wb, 'choices')
        self.ext_choices = self.get_choices(wb, 'external_choices')
        self.metadata = {
            **self.metadata,
            **{
                'file_name': os.path.split(wb.file)[1],
                'form_id': self.settings.get('form_id') if self.settings.get('form_id') else self.settings.get('id_string'),
                'country': lambda:self.metadata['form_id'][3:5],
                'round': lambda:self.metadata['form_id'][6:7],
                'type_of_form': lambda:self.metadata['form_id'][0:2],
            }
        }
        self.metadata = {
            **self.metadata,
            'country': self.metadata['country'](),
            'round': self.metadata['round'](),
            'type_of_form': self.metadata['type_of_form'](),
        }
        self.questionnaire = self.convert_survey(wb, self.choices,
                                                 self.ext_choices)

    unhandled_token_types = \
        ['calculate', 'start', 'end', 'deviceid', 'simserial',
         'phonenumber', 'hidden', 'hidden string', 'hidden int',
         'hidden geopoint']
    warnings = None

    @classmethod
    def from_file(cls, path):
        """Create Odkform object from file in path.

        Args:
            path (str): The path for the source file of the ODK form,
                typically an '.xlsx' file meeting the XLSForm specification.

        Returns:
            Odkform
        """
        xlsform = Xlsform(path)

        return cls(xlsform)

    @staticmethod
    def get_settings(wb):
        """Get the XLSForm settings as a settings_dict.

        Args:
            wb (Xlsform): A workbook object representing ODK form.

        Returns:
            dict: Form settings.
        """
        settings_dict = {}
        try:
            settings = wb['settings']
            header = settings[0]
            details = settings[1]
            settings_dict = {k: v for k, v in zip(header, details)}
        except (KeyError, IndexError):
            # KeyError: Worksheet does not exist.
            # IndexError: Does not have the correct rows.
            pass
        return settings_dict

    @staticmethod
    def get_choices(wb, ws):
        """Extract choices from an XLSForm.

        Args:
            wb (Xlsform): A Xlsform object representing ODK form.
            ws (Worksheet): One of 'choices' or 'external_choices'.

        Returns:
            dict: A dictionary of choice list names with list of choices
                options for each list.

        Raises:
            OdkformError: Catches instances where list specified in the
                'survey' worksheet, but the list does not appear in the
                designated 'choices' or 'external_choices' worksheet.
        """
        formatted_choices = {}
        try:
            choices = wb[ws]
            header = [str(x) for x in choices[0]]

            if 'list_name' not in header:
                msg = 'Column "list_name" not found in {} tab'.format(ws)
                raise OdkFormError(msg)

            for i, row in enumerate(choices):
                if i == 0:
                    continue
                dict_row = {str(k): str(v) for k, v in zip(header, row)}
                list_name = dict_row['list_name']
                if list_name in formatted_choices:
                    formatted_choices[list_name].add(dict_row)
                elif list_name:  # Not "else:" because possibly blank rows.
                    odkchoices = OdkChoices(list_name)
                    odkchoices.add(dict_row)
                    formatted_choices[list_name] = odkchoices
        except (KeyError, IndexError):  # Worksheet does not exist.
            pass
        return formatted_choices

    @staticmethod
    def get_title(settings, wb, lang=None):
        """Get questionnaire title.

        Args:
        settings (dict): A dictionary represetnation of the original 'settings'
            worksheet of an ODK XLSForm.
        wb (Workbook): A Workbook object representing an XLSForm.
        lang (str): The requested render language of the form.

        Returns:
            str: The title.
        """
        lookup_title = 'form_title'
        if lang:
            try1 = settings.get('ppp_form_title' + '::' + lang)
            try2 = settings.get('ppp_form_title' + ':' + lang)
            lookup_title = try1 if try1 else try2
        backup_title = os.path.split(wb.file)[1]
        return settings.get(lookup_title, backup_title)

    def to_text(self, lang=None):
        """Get the text representation of an entire XLSForm.

        Args:
            lang (str): The language.

        Returns:
            str: The full string of the XLSForm, ready to print or save.
        """
        title_lines = (
            '+{:-^50}+'.format(''),
            '|{:^50}|'.format(self.title),
            '+{:-^50}+'.format('')
        )
        title_box = '\n'.join(title_lines)

        q_text = (q.to_text(lang) for q in self.questionnaire)
        sep = '\n\n' + '=' * 52 + '\n\n'
        result = sep.join(q_text)
        return title_box + sep + result + sep  # TODO: Finish below to_dict or
    # TODO: change debug feature. If fixed, change to_json to
    # TODO: call dump return of this method instead of raw data.

    # def to_dict(self, lang):
    #     """Get the dictionary representation of an entire XLSForm.
    #
    #     Args:
    #         lang (str): The language.
    #
    #     Returns:
    #         dict: A full dictionary representation of the XLSForm.
    #
    #     """
    #     lang = lang if lang \
    #         else self.languages['language_list']['default_language']
    #     html_questionnaire = {
    #         'title': self.title,
    #         'questions': []
    #     }
    #     for item in self.questionnaire:
    #         html_questionnaire['questions'].append(item.to_dict(lang))
    #     return html_questionnaire

    def to_json(self, pretty=False):
        """Get the JSON representation of raw ODK form.

        Args:
            pretty (bool): Activates prettification, involving insertion of
                several kinds of whitespace for readability.

        Returns:
            json: A full JSON representation of the XLSForm.
        """
        import json
        raw_survey = []
        header = self.metadata['raw_data']['survey'][0]
        for i, row in enumerate(self.metadata['raw_data']['survey']):
            if i == 0:
                continue
            raw_survey.append({str(k): str(v) for k, v in zip(header, row)})

        if pretty:
            return json.dumps(raw_survey, indent=2)
        return json.dumps(raw_survey)

    @staticmethod
    def _add_i_nums_to_questions(obj, data={'qnum':'', 'i':0}, depth=0):
        """Add iteration numbers to unique question prompts.

        Possible improvements: Make it so that this doesn't depend on
        OdkPrompt._extract_question_numbers.

        Args:
            obj (list): From either: OdkForm.questionnaire, OdkGroup.row, or
            OdkRepeat.row.
            data (dict): Running data. Tracks current question number 'qnum'
            and also current iteration 'i'. The 'i' should increment by 1 for
            each new question number. In this case, a question number is
            rigidly considered "new" if it is not exactly equal to whatever
            was considered to be the immediately preceding question number
            in the form.
            depth (int): Recursion depth. Starts at 0. Increments by 1 each
            time this function recurses, which happens every time an OdkRepeat
            or OdkGroup is encountered.

        Returns:
            list: OdkForm.questionnaire with new 'i' values included.
        """
        for i, element in enumerate(obj):
            if isinstance(element, OdkGroup) or isinstance(element, OdkRepeat):
                element.data, data = OdkForm._add_i_nums_to_questions(
                    element.data, data, depth+1)
            elif isinstance(element, OdkPrompt):
                element.row = OdkPrompt._extract_question_numbers(element.row)
                qnum = element.row['question_number']
                if qnum not in (data['qnum'], ''):
                    data['i'] += 1
                    data['qnum'] = qnum
                element.row['i'] = data['i']
            elif isinstance(element, OdkCalculate):
                element.row['i'] = data['i']
            elif isinstance(element, OdkTable):  # TODO
                pass
            else:
                pass
        if depth == 0:
            return obj
        return obj, data

    def to_html(self, lang=None, **kwargs):
        """Get the JSON representation of an entire XLSForm.

        Args:
            lang (str): The language.
            **highlight (bool): For color highlighting of various components
                of html template.
            **debug (bool): For inclusion of debug information to be printed
                in the JavaScript console.

        Returns:
            str: A full HTML representation of the XLSForm.
        """
        language = lang if lang else self.language
        debug = True if 'debug' in kwargs and kwargs['debug'] else False
        html_questionnaire = ''
        data = {
            'header': {
                'title': self.get_title(settings=self.settings,
                                        wb=self.metadata['raw_data'],
                                        lang=lang)
            },
            'footer': {
                'data': self.to_json(pretty=True) if debug else 'false'
            },
            'questionnaire': OdkForm._add_i_nums_to_questions(
                self.questionnaire)
        }

        # - Render Header
        # pylint: disable=no-member
        header = TEMPLATE_ENV.get_template('header.html')\
            .render(data=data['header'], **kwargs)
        # pylint: disable=no-member
        grp_spc = TEMPLATE_ENV\
            .get_template('content/group/group-spacing.html').render()
        html_questionnaire += header

        # - Render Body
        prev_item = None
        hlt = kwargs['highlight'] if 'highlight' in kwargs else False
        for index, item in enumerate(data['questionnaire']):

            if exclusion(item=item, settings=kwargs):
                continue

            if prev_item is not None and isinstance(item, OdkGroup):
                html_questionnaire += grp_spc
            elif isinstance(prev_item, OdkGroup) \
                    and not isinstance(item, OdkGroup):
                html_questionnaire += grp_spc
            if isinstance(item, OdkPrompt) and item.is_section_header and \
                    isinstance(data['questionnaire'][index+1], OdkGroup):
                html_questionnaire += \
                    item.to_html(language, hlt, **kwargs, bottom_border=True)
            else:
                html_questionnaire += item.to_html(language, hlt, **kwargs)
            prev_item = item
        OdkForm.warnings = OdkForm.warnings if OdkForm.warnings else 'false'

        # pylint: disable=no-member
        footer = TEMPLATE_ENV.get_template('footer.html')\
            .render(info=None, warnings=OdkForm.warnings,
                    data=data['footer']['data'], **kwargs)
        html_questionnaire += footer

        return html_questionnaire

    @staticmethod
    def parse_select_type(row, choices, ext_choices):
        """Extract relevant information from a select_* ODK prompt.

        Build a dictionary that distills the main details of the row. The
        select type questions can have a token type of 'prompt' or 'table'.
        The prompt type is default, and table type is if the appearance of the
        question has either 'label' or 'list-nolabel'.

        Args:
            row (dict): A row as a dictionary. Keys and values are strings.
            choices (dict): A diction   ary with list_names as keys. Represents
                the choices found in 'choices' tab.
            ext_choices (dict): A dictionary with list_names as keys.
                Represents choices found in 'external_choices' tab.

        Returns:
            A dictionary with the simple information about this prompt.

        Raises:
            OdkFormError: If the row is not select_[one|multiple](_external)?
            KeyError: If the select question's choice list is not found.
        """
        simple_row = {'token_type': 'prompt'}
        simple_type = 'select_one'
        row_type = row['type']
        list_name = row_type.split(maxsplit=1)[1]

        try:
            if row_type.startswith('select_one_external '):
                choice_list = ext_choices[list_name]
            elif row_type.startswith('select_multiple_external '):
                simple_type = 'select_multiple'
                choice_list = ext_choices[list_name]

            elif row_type.startswith('select_one '):
                choice_list = choices[list_name]
            elif row_type.startswith('select_multiple '):
                simple_type = 'select_multiple'
                choice_list = choices[list_name]
            else:
                raise OdkFormError()
        except KeyError:
            raise OdkFormError('List \'{}\' not found.'.format(list_name))

        simple_row['simple_type'] = simple_type
        simple_row['choice_list'] = choice_list

        appearance = row.get('appearance', '')
        if 'label' in appearance or 'list-nolabel' in appearance:
            simple_row['token_type'] = 'table'

        return simple_row

    @staticmethod
    def parse_group_repeat(row):
        """Extract relevant information about a begin/end group/repeat.

        Args:
            row (dict): A row as a dictionary. Keys and values are strings.

        Returns:
            A dictionary with the simple information about this prompt.

        Raises:
            OdkFormError: If type is not begin/end group/repeat.
        """
        row_type = row['type']
        token_type = row_type
        appearance = row.get('appearance', '')
        good = ('begin group', 'end group', 'begin repeat', 'end repeat')
        if row_type == 'begin group' and 'field-list' not in appearance:
            token_type = 'context group'
        elif row_type not in good:
            raise OdkFormError()
        simple_row = {'token_type': token_type}
        return simple_row

    @staticmethod
    def make_simple_prompt(row_type):
        """Extract relevant information from an ODK prompt.

        Make the simplest dictionary: token_type is set to 'prompt',
        simple_type is copied from the row type, and choices is set to None.

        Args:
            row_type (str): The type of the row.

        Returns:
            A dictionary with the simple information about this prompt.
        """
        simple_row = {
            'token_type': 'prompt',
            'simple_type': row_type,
            'choice_list': None
        }
        return simple_row

    @staticmethod
    def make_simple_calculate():
        """Make a simple calculate."""
        simple_row = {
            'token_type': 'calculate'
        }
        return simple_row

    @staticmethod
    def parse_type(row, choices, ext_choices):
        """Describe the 'type' column of a row XLSForm.

        Args:
            row (dict): A row as a dictionary. Keys and values are strings.
            choices (dict): A dictionary with list_names as keys. Represents
                the choices found in 'choices' tab.
            ext_choices (dict): A dictionary with list_names as keys.
                Represents choices found in 'external_choices' tab.

        Returns:
            dict: simple_row information from parsing.
        """
        row_type = row['type']
        simple_types = OdkPrompt.response_types + OdkPrompt.non_response_types
        if row_type in simple_types:
            simple_row = OdkForm.make_simple_prompt(row_type)
        elif row_type == 'calculate':
            simple_row = OdkForm.make_simple_calculate()
        elif row_type.startswith('select_'):
            simple_row = OdkForm.parse_select_type(row, choices, ext_choices)
        elif row_type.startswith('begin ') or row_type.startswith('end '):
            simple_row = OdkForm.parse_group_repeat(row)
        else:  # Note - Some unhandled token types remain.
            simple_row = {'token_type': 'unhandled', 'simple_type': row_type}
        return simple_row

    # pylint: disable=too-many-branches
    @staticmethod
    def convert_survey(wb, choices, ext_choices):
        """Convert rows and strings of a workbook into object components.

        Main types are:

        - prompt
        - calculate
        - begin group
        - end group
        - begin repeat
        - end repeat
        - table
        - context group (group without field-list appearance)

        Args:
            wb (Xlsform): A Xlsform object representing an XLSForm.

        Returns:
            list: A list of objects representing form components.

        Raises:
            OdkformError: Handle several errors, including: mismatched groups
                or repeat groups, errors when appending to groups or repeat
                groups, erroneously formed tables, duplicate context group
                names, and groups nested within a field-list group.
        """
        context = OdkForm.ConversionContext()
        survey, header = None, None
        try:
            survey = wb['survey']
            header = survey[0]
        except KeyError:  # No survey found.
            pass

        if survey and header:
            for i, row in enumerate(survey):
                if i == 0:
                    continue
                dict_row = {str(k): str(v) for k, v in zip(header, row)}
                token = OdkForm.parse_type(dict_row, choices, ext_choices)

                if token['token_type'] == 'prompt':
                    dict_row['simple_type'] = token['simple_type']
                    choice_list = token['choice_list']
                    this_prompt = OdkPrompt(dict_row, choice_list)
                    context.add_prompt(this_prompt)
                elif token['token_type'] == 'calculate':
                    this_calculate = OdkCalculate(dict_row)
                    context.add_calculate(this_calculate)
                elif token['token_type'] == 'begin group':
                    this_group = OdkGroup(dict_row)
                    context.add_group(this_group)
                elif token['token_type'] == 'context group':
                    # Possibly make an OdkGroup here...
                    context.add_context_group()
                elif token['token_type'] == 'end group':
                    context.end_group()
                elif token['token_type'] == 'begin repeat':
                    this_repeat = OdkRepeat(dict_row)
                    context.add_repeat(this_repeat)
                elif token['token_type'] == 'end repeat':
                    context.end_repeat()
                elif token['token_type'] == 'table':
                    dict_row['simple_type'] = token['simple_type']
                    choice_list = token['choice_list']
                    this_prompt = OdkPrompt(dict_row, choice_list)
                    context.add_table(this_prompt)
                elif token['token_type'] == 'unhandled':
                    # Intentionally no handling for these types.
                    # TODO: Possibly start logging?
                    pass
                else:
                    # TODO: Make an error?
                    pass

        return context.result

    class ConversionContext:
        """A class to help questionnaire conversion.

        This class is the context during questionnaire conversion. It
        remembers state, adds components in the correct order, and enforces
        rules during conversion.

        Instance attributes:
            result (list): The list of survey components that is built up
            pending_stack (list): A stack for tracking nested groups and
                repeats.
            group_stack (list): A stck for tracking nested groups and context
                groups.
        """

        def __init__(self):
            """Initialize a conversion context before parsing."""
            self.result = []
            self.pending_stack = []
            self.group_stack = []

        def add_prompt(self, prompt):
            """Add a prompt to the questionnaire.

            If there is an item on the pending stack, it is added there,
            otherwise it is added to the list of components.

            Args:
                prompt (OdkPrompt | OdkCalculate): A prompt to add.

            """
            if self.pending_stack:
                self.pending_stack[-1].add(prompt)
            else:
                self.result.append(prompt)

        def add_calculate(self, calculate):
            """Add a calculate to the questionnaire.

            If there is an item on the pending stack, it is added there,
            otherwise it is added to the list of components.

            Args:
                calculate (OdkCalculate): A calculate to add.

            """
            self.add_prompt(calculate)

        def add_group(self, group):
            """Add a group to the pending stack.

            A group can be added to the pending stack as long as it is empty
            or the last pending stack item is a repeat. This is triggered by a
            'begin group' row with a 'field-list' in the appearance.

            Args:
                group (OdkGroup): The group to add to the pending stack.

            Raises:
                OdkFormError: If the parsing rules are broken based on the
                    current context.

            """
            if self.pending_stack:
                last = self.pending_stack[-1]
                if isinstance(last, OdkGroup):
                    msg = 'Groups cannot be nested in each other.'
                    raise OdkFormError(msg)
            self.pending_stack.append(group)
            self.group_stack.append(group)

        def add_context_group(self):
            """Add a context group to the group stack.

            Context groups are tracked only to help popping groups correctly
            from the pending stack.

            """
            self.group_stack.append(None)

        def end_pending_group(self):
            """End a pending group.

            A pending group is a group on the pending stack. This is not a
            context group. This function is only called internally in response
            to receiving and dealing with an 'end group' type.

            If the pending group is nested in a repeat, then it is added to
            that repeat.

            Raises:
                OdkFormError: If the parsing rules are broken based on the
                    current context.

            """
            if self.pending_stack:
                last_pending = self.pending_stack.pop()
                if not isinstance(last_pending, OdkGroup):
                    msg = 'Found end group but no group in pending stack'
                    raise OdkFormError(msg)
                last_pending.add_pending()
                if self.pending_stack:
                    self.pending_stack[-1].add(last_pending)
                else:
                    self.result.append(last_pending)
            else:
                msg = 'Found end group but nothing pending stack.'
                raise OdkFormError(msg)

        def end_group(self):
            """Finish a group after seeing 'end group' type.

            The 'end group' type can finish a field-list group or a context
            group. This function handles the logic for finishing both types.

            Raises:
                OdkFormError: If the parsing rules are broken based on the
                    current context.

            """
            if self.group_stack:
                last_group = self.group_stack.pop()
                if isinstance(last_group, OdkGroup):
                    self.end_pending_group()
            else:
                msg = 'Begin/end group mismatch'
                raise OdkFormError(msg)

        def add_repeat(self, repeat):
            """Add a repeat to the pending stack.

            The pending stack must first be empty because a repeat cannot be
            nested in a group or other repeat.

            Args:
                repeat (OdkRepeat): The repeat to deal with.

            Raises:
                OdkFormError: If the parsing rules are broken based on the
                    current context.

            """
            if not self.pending_stack:
                self.pending_stack.append(repeat)
            else:
                msg = 'Unable to nest repeat inside a group or repeat.'
                raise OdkFormError(msg)

        def end_repeat(self):
            """Finish a repeat in this questionniare.

            A repeat can be ended only if it is first on the pending stack.

            Raises:
                OdkFormError: If the parsing rules are broken based on the
                    current context.

            """
            if self.pending_stack:
                last_pending = self.pending_stack.pop()
                if isinstance(last_pending, OdkRepeat):
                    self.result.append(last_pending)
                else:
                    msg = 'Found end repeat but no repeat in pending stack.'
                    raise OdkFormError(msg)
            else:
                msg = 'Found end repeat but nothing in pending stack.'
                raise OdkFormError(msg)

        def add_table(self, prompt):
            """Add a table row to the questionnaire.

            The table can only be added if there is a group on the pending
            stack.

            Args:
                prompt (OdkPrompt): The prompt representing the table row.

            Raises:
                OdkFormError: If the parsing rules are broken based on the
                    current context.

            """
            if self.pending_stack:
                last_pending = self.pending_stack[-1]
                if not isinstance(last_pending, OdkGroup):
                    msg = 'A table can only be in a group.'
                    raise OdkFormError(msg)
                last_pending.add_table(prompt)
            else:
                msg = 'A table can only be in a group, no group found.'
                raise OdkFormError(msg)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
About PPP
PPP is a tool that converts ODK XlsForm specification .xlsx Excel files to more
human-readable, printable formats, commonly called "paper questionnaires".

Officially, PPP stands for "Pretty PDF Printer", but other formats are
supported (see: the `--format` option described in the "CLI How-to").

PPP Web Application
- Convert XlsForms online: http://ppp.pma2020.org
- Github: https://github.com/pma-2020/ppp-web

Functions
- run: Common executional entry point from interfaces.
"""
import os
from copy import copy
try:
    # noinspection PyUnresolvedReferences
    from signal import signal, SIGPIPE, SIG_DFL
except ImportError as e:
    pass
from itertools import product
from collections import OrderedDict

from ppp.definitions.error import OdkException, InvalidLanguageException
from ppp.definitions.constants import MULTI_ARGUMENT_CONVERSION_OPTIONS
from ppp.odkform import OdkForm, set_template_env


def convert_file(in_file, language=None, outpath=None, **kwargs):
    """Run ODK form conversion.

    Args:
        in_file (str): Path to load source file.
        language (str or None): Language to render form.
        outpath (str or None): Path to save converted file.
        **format (str): File format to be output.
        **debug (bool): Debugging on or off.
        **highlight (bool): Highlighting on or off.

    Raises:
        InvalidLanguageException: Language related.
        OdkChoicesError: Choice or choice list related.
        OdkFormError: General form related exception.
    """
    
    set_template_env(kwargs['template'] if 'template' in kwargs else 'default')

    form = OdkForm.from_file(in_file)

    try:
        output = None
        output_format = \
            kwargs['format'] if 'format' in kwargs else 'html'
        if output_format == 'text':
            output = form.to_text(lang=language, **kwargs)

        elif output_format in ('html', 'doc'):
            output = form.to_html(lang=language, **kwargs)

        if outpath:
            if os.path.isdir(outpath) and not os.path.exists(outpath):
                os.makedirs(outpath)
            if os.path.isdir(outpath):
                base_filename = os.path.basename(os.path.splitext(in_file)[0])
                lang = '-' + language if language else ''
                options_affix = '-' + kwargs['preset'] \
                    if 'preset' in kwargs and kwargs['preset'] != 'developer' \
                    else ''
                out_file = '{}{}{}{}.{}'.format(outpath, base_filename, lang,
                                                options_affix, output_format)

                if isinstance(out_file, list):
                    if out_file[0] == '/':
                        out_file = out_file[1:]
            else:
                out_file = outpath
            with open(out_file, mode='w', encoding='utf-8') as file:
                file.write(output)
            print(out_file)
        else:
            try:
                print(output)
            except BrokenPipeError:  # If output is piped.
                try:
                    signal(SIGPIPE, SIG_DFL)
                # noinspection PyBroadException
                except:
                    pass
                print(output)
    except InvalidLanguageException as err:
        if str(err):
            raise InvalidLanguageException(err)
        elif language is None:
            msg = 'InvalidLanguageException: An unknown error occurred when ' \
                  'attempting to convert form. If a language was not ' \
                  'supplied, please supply and try again.'
            raise InvalidLanguageException(msg)
    except OdkException as err:
        raise OdkException(err)


def enumerate_combos(dict_with_lists):
    """Enumerate keyword-arg combination variants.

    Args:
        dict_with_lists (dict): A dictionary of lists.

    Returns:
        list: A list of dicts, where each list in dict_of_lists has been
        de-listed to a single value.
    """
    dict_with_only_lists = \
        OrderedDict({k: v for k, v in dict_with_lists.items()
                     if isinstance(v, list)})
    dict_without_lists = \
        {k: v for k, v in dict_with_lists.items() if not isinstance(v, list)}

    ds1 = dict_with_only_lists.values()
    ds2 = [i for i in ds1]
    ds3 = list(product(*ds2))
    ds4 = [dict(zip(dict_with_only_lists, i)) for i in ds3]
    full_dict_combos = [{**i, **dict_without_lists} for i in ds4]

    return full_dict_combos


def num_args(option):
    """Get number of args, aka "nargs" for a given option."""
    return len(option) if isinstance(option, list) else 1


def run(files, languages=[None], outpath=None, **kwargs):
    """Run ODK form conversion on n files of n option combinations.

    Args:
        files (list): Path to load source file.
        languages (list): Languages to render forms.
        output_format (str): File format to be output.
        outpath (str): Path of file name to save converted file if 1 file,
            else path to directory for multiple files, in which case file names
            will be automatically generated.
        **debug (bool): Debugging on or off.
        **highlight (bool): Highlighting on or off.
    """
    _outpath = outpath
    _kwargs = copy(kwargs)
    combos = enumerate_combos(_kwargs)
    num_output = num_args(files) * num_args(languages) * num_args(combos)

    if num_output > 1 or outpath:
        print('Creating files.')

    for file in files:
        if num_output > 1 and not outpath:
            _outpath = os.path.dirname(file) + '/'
        for language in languages:
            for combo in combos:
                convert_file(file, language, outpath=_outpath, **combo)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Unit tests for PPP package."""
import unittest
import os
import subprocess
from glob import glob

from ppp.odkform import OdkForm, set_template_env
from ppp.odkprompt import OdkPrompt
from ppp.odkgroup import OdkGroup
from test.config import TEST_FILES_DIR, TEST_PACKAGES
from test.utils import get_args, get_test_suite
# from pmix.odkchoices import OdkChoices  # TODO
# from pmix.odkrepeat import OdkRepeat  # TODO
# from pmix.odktable import OdkTable  # TODO


# # Mock Objects
class MockForm(OdkForm):
    """Mock object of OdkForm"""

    def __init__(self, mock_file, mock_dir=None):
        """Initialize mock form.

        Args:
            mock_file (str): File name.
            mock_dir (str): Directory.
        """
        path = mock_dir + mock_file if mock_dir else TEST_FILES_DIR + mock_file
        form = super().from_file(path)
        super().__init__(form)


# # Unit Tests
# pylint: disable=too-few-public-methods
# - PyLint check not apply? - http://pylint-messages.wikidot.com/messages:r0903
class PppTest(unittest.TestCase):
    """Base class for PPP package tests."""

    @classmethod
    def files_dir(cls):
        """Return name of test class."""
        return TEST_FILES_DIR + cls.__name__

    def input_path(self):
        """Return path of input file folder for test class."""
        return self.files_dir() + '/input/'

    def output_path(self):
        """Return path of output file folder for test class."""
        return self.files_dir() + '/output/'

    def input_files(self):
        """Return paths of input files for test class."""
        all_files = glob(self.input_path() + '*')
        # With sans_temp_files, you can have XlsForms open while testing.
        sans_temp_files = [x for x in all_files
                           if not x[len(self.input_path()):].startswith('~$')]
        return sans_temp_files

    def output_files(self):
        """Return paths of input files for test class."""
        return glob(self.output_path() + '*')

    def standard_convert(self):
        """Converts input/* --> output/*. Returns n files each."""
        in_files = self.input_files()
        out_dir = self.output_path()

        subprocess.call(['rm', '-rf', out_dir])
        os.makedirs(out_dir)
        command = ['python3', '-m', 'ppp'] + in_files + ['-o', out_dir]
        subprocess.call(command)

        expected = 'N files: ' + str(len(in_files))
        actual = 'N files: ' + str(len(self.output_files()))
        return expected, actual

    def standard_conversion_test(self):
        """Checks standard convert success."""
        expected, actual = self.standard_convert()
        self.assertEqual(expected, actual)

    @staticmethod
    def get_forms(data):
        """Convert specified forms from form name strings to objects."""
        forms = {}
        for datum in data:
            # Should streamline setUps. Currently in both tuple and dict.
            try:
                file = datum[0]['file']
            except KeyError:
                file = datum['inputs']['file']
            if file not in forms:
                forms[file] = OdkForm.from_file(TEST_FILES_DIR + file)
        return forms


class OdkPromptTest(PppTest):
    """Unit tests for the OdkPrompt class."""

    media_types = ['image', 'audio', 'video', 'media::image',
                   'media::audio', 'media::video']
    media_lead_char = '['
    media_end_char = ']'
    arbitrary_language_param = 'English'

    def setUp(self):
        """Set up."""
        self.data = (
            ({'inputs': {
                'file': 'FQ.xlsx'
            }, 'expected_outputs': {

            }},
             {'inputs': {
                 'file': 'HQ.xlsx'
             }, 'expected_outputs': {

             }},
            )
        )

    def test_to_dict(self):
        """Test to_dict method."""
        def test_media_fields_in_prompts():
            """Tests for media fields."""
            def asserts(item_dict):
                """Iterate through asserts."""
                for key, val in item_dict.items():
                    for media_type in OdkPromptTest.media_types:
                        if key.startswith(media_type) and val:
                            # A field such as 'media::image::English'
                            # is formatted correctly.
                            self.assertTrue(val[0] == lead_char and
                                            val[-1] == end_char)
                            # A field such as 'image' exists and is
                            # formatted correctly.
                            self.assertTrue(
                                item_dict[media_type][0] == lead_char and
                                item_dict[media_type][-1] == end_char)
                            # No discrepancies between language based and non
                            # language based media fields.
                            self.assertTrue(item_dict[media_type] == val)
                            # The field 'media' exists and formatted correct.
                            self.assertTrue(item_dict['media'])
            lang = OdkPromptTest.arbitrary_language_param
            lead_char = OdkPromptTest.media_lead_char
            end_char = OdkPromptTest.media_end_char
            forms = self.get_forms(self.data)
            for i in self.data:
                file_name = i['inputs']['file']
                for item in forms[file_name].questionnaire:
                    if isinstance(item, OdkPrompt):
                        asserts(item.to_dict(lang))

        test_media_fields_in_prompts()

    def test_initialization_has_choices(self):
        """Test that choice list exists on initialization."""
        forms = self.get_forms(self.data)
        for dummy, form in forms.items():
            for item in form.questionnaire:
                if isinstance(item, OdkPrompt):
                    if item.odktype in item.select_types:
                        msg = 'No choices found in \'{}\'.'.format(item)
                        self.assertTrue(item.choices is not None, msg=msg)

    def test_question_number_field_from_form(self):
        """Test that question numbers are geing generated as expected."""
        expected_outputs = {
            'ever_birth': '200',
            'fb_note': '205',
            'birth_events_yes': 'LCL201',
            'children_living': '201a'
        }
        test_language = 'English'
        forms = []
        form_names = ['OdkFormTest', 'QuestionNumberExtractionTest']
        for form in form_names:
            forms.append(OdkForm.from_file(TEST_FILES_DIR + form + '.xlsx'))
        for form in forms:
            for item in form.questionnaire:
                if isinstance(item, OdkPrompt):
                    row = item.to_dict(lang=test_language)
                    if row['name'] in expected_outputs:
                        msg = 'Question number \'{}\' did not match what was '\
                              'expected for the following question: \n\n{}'\
                            .format(row['question_number'], row['label'])
                        self.assertTrue(row['question_number'] ==
                                        expected_outputs[row['name']], msg=msg)

    def test_question_number_field_statically(self):
        """Test that question numbers are geing generated as expected."""
        inputs_to_outputs = {
            '201. Random question': '201',
            'LCL_202. Random question': 'LCL_202',
            'CRVS-203. Random question': 'CRVS-203',
            '204.A. Random question': '204.A',
            '205.1. Random question': '205.1',
            '206.a. Random question': '206.a',
            '207.i. Random question': '207.i',
            '208.C.ii.1. Random question': '208.C.ii.1',
            '209.C.ii.1_v2. Random question': '209.C.ii.1_v2',
            '210.C.ii.1_v2-2. Random question': '210.C.ii.1_v2-2',
            'State: ${level1_unlinked}': ''
        }
        for k, v in inputs_to_outputs.items():
            prompt = {
                'label': k
            }
            question_number = \
                OdkPrompt._extract_question_numbers(prompt)['question_number']
            msg = 'Question number \'{}\' did not match what was ' \
                  'expected for the following question: \n\n{}'\
                .format(question_number, k)
            self.assertTrue(question_number == v, msg=msg)


class OdkGroupTest(unittest.TestCase):
    """Unit tests for the OdkGroup class."""

    def setUp(self):
        """Set up."""
        self.data = (
            ({'inputs': {'type': 'begin group', 'name': 'date_group',
                         'label::English': '', 'hint::English': '',
                         'constraint_message::English': '', 'constraint': '',
                         'required': '', 'appearance': 'field-list',
                         'default': '',
                         'relevant': 'today() > date("2017-03-01") and '
                                     'today() < date("2017-11-01")',
                         'read_only': '', 'calculation': '', 'choice_filter':
                             '', 'image::English': '',
                         'save_instance': '', 'save_form': '', 'label::Ateso':
                             '', 'label::Luganda': '',
                         'label::Lugbara': '', 'label::Luo': '',
                         'label::Lusoga': '', 'label::Ngakarimojong': '',
                         'label::Runyankole-Rukiga': '',
                         'label::Runyoro-Rutoro': '', 'hint::Ateso': '',
                         'hint::Luganda': '', 'hint::Lugbara': '', 'hint::Luo':
                             '', 'hint::Lusoga': '',
                         'hint::Ngakarimojong': '', 'hint::Runyankole-Rukiga':
                             '', 'hint::Runyoro-Rutoro': '',
                         'constraint_message::Ateso': '',
                         'constraint_message::Luganda': '',
                         'constraint_message::Lugbara': '',
                         'constraint_message::Luo': '',
                         'constraint_message::Lusoga': '',
                         'constraint_message::Ngakarimojong': '',
                         'constraint_message::Runyankole-Rukiga': '',
                         'constraint_message::Runyoro-Rutoro': '',
                         'label': '', 'hint': '', 'constraint_message': '',
                         'image': '', 'constraint_original': '',
                         'relevant_original': 'today() > date("2017-03-01") '
                                              'and today() < date("2017-11-01"'
                                              ')',
                         'input_field': None},
              'outputs': {
                  'header_primitive_type': type({})
              }}),
        )

    def test_render_header(self):
        """Test header rendering."""
        for i in self.data:
            expected_output = i['outputs']['header_primitive_type']
            output = type(OdkGroup(i['inputs']).format_header(i['inputs']))
            msg = '\n\nCase:\n{}\nOutput:\n{}\nExpected:\n{}'\
                .format(i, output, expected_output)
            self.assertTrue(output == expected_output, msg=msg)


class OdkFormTest(PppTest):
    """Unit tests for the OdkForm class."""

    def setUp(self):
        """Set up."""
        self.data = (
            {'inputs': {'file': 'OdkFormTest.xlsx'}, 'position': 0,
             'outputs': {'class': OdkPrompt,
                         'repr': '<OdkPrompt ever_birth>'}},
            {'inputs': {'file': 'OdkFormTest.xlsx'}, 'position': 1,
             'outputs': {'class': OdkGroup,
                         'repr': '<OdkGroup FB: [<OdkPrompt fb_note>, '
                                 '<OdkPrompt fb_m>, <OdkPrompt fb_y>]>'}},
            {'inputs': {'file': 'OdkFormTest.xlsx'}, 'position': 2,
             'outputs': {'class': OdkPrompt,
                         'repr': '<OdkPrompt birth_events_yes>'}},
            {'inputs': {'file': 'OdkFormTest.xlsx'}, 'position': 3,
             'outputs': {'class': OdkPrompt,
                         'repr': '<OdkPrompt children_living>'}},
        )

    def test_questionnaire(self):
        """Test expected results of converted questionnaire based on position.
        """
        forms = self.get_forms(self.data)

        for datum in self.data:
            expected_output = datum['outputs']
            output = \
                forms[datum['inputs']['file']].questionnaire[datum['position']]

            # - Check Object Representation
            got = str(output)
            expected = expected_output['repr']
            msg = '\nGot: {}\nExpected: {}'.format(got, expected)
            self.assertEqual(got, expected, msg)

            # - Check Object Class
            got = output
            expected = expected_output['class']
            msg = '\nGot: {}\nExpected: {}'.format(got, expected)
            # noinspection PyTypeChecker
            self.assertTrue(isinstance(got, expected), msg)

    def test__add_prompt_iteration_numbers_to_form(self):
        """Test _add_prompt_iteration_numbers_to_form method.

            This test pulls from uses a particular file which has had a special
            column added to it called "i". For any row in the form file which
            has data, there is a value for "i" that has been pre-set to what
            we would expect.
        """
        rel_file_path = 'test__add_prompt_iteration_numbers_to_form/1.xlsx'
        form = OdkForm.from_file(TEST_FILES_DIR + rel_file_path)
        with_iterations = form._add_i_nums_to_questions(form.questionnaire)
        expected = 137
        actual = int(with_iterations[-1].row['i'])
        msg = 'Expected iteration {} does not match actual value of {}.'\
            .format(expected, actual)
        self.assertTrue(expected == actual, msg)

    def test_to_html(self):
        """Test to_html method."""
        set_template_env('old')
        forms = self.get_forms(self.data)
        for dummy, form in forms.items():
            kwargs = {'format': 'html'}
            html = form.to_html(**kwargs)
            self.assertTrue(html.startswith("""<html>\n    <head>"""))
            self.assertTrue(html.endswith('</html>'))
            self.assertIn("""    </head>\n    <body id="survey">
        <div id="container">
            <div id="identification_section" class="section">""", html)


class MultiConversionTest(unittest.TestCase):
    """Test conversion of n files in n languages for n option combinations."""

    maxDiff = None  # Allows to see full error output for this test.

    def test_multi_conversion(self):
        src_dir = TEST_FILES_DIR + 'multiple_file_language_option_conversion/'
        out_dir = src_dir + 'ignored/'
        src_dir_ls_input = os.listdir(src_dir)
        src_files = \
            [src_dir + x for x in src_dir_ls_input if x.endswith('.xlsx')]
        subprocess.call(['rm', '-rf', out_dir])
        os.makedirs(out_dir)
        command = ['python3', '-m', 'ppp'] + src_files + \
                  ['-o', out_dir, '-f', 'doc', 'html', '-p', 'minimal', 'full',
                   '-l', 'English', 'Français']
        subprocess.call(command)

        out_dir_ls_input_unsorted = os.listdir(out_dir)

        expected_output_unsorted = \
            ['BFR5-Selection-v2-jef-Français-minimal.html',
             'BFR5-Female-Questionnaire-v13-jef-Français-full.html',
             'BFR5-Female-Questionnaire-v13-jef-English-minimal.doc',
             'BFR5-Female-Questionnaire-v13-jef-Français-full.doc',
             'BFR5-Selection-v2-jef-Français-full.html',
             'BFR5-Female-Questionnaire-v13-jef-English-full.doc',
             'BFR5-Female-Questionnaire-v13-jef-Français-minimal.html',
             'BFR5-Selection-v2-jef-Français-full.doc',
             'BFR5-Female-Questionnaire-v13-jef-Français-minimal.doc',
             'BFR5-Female-Questionnaire-v13-jef-English-full.html',
             'BFR5-Selection-v2-jef-English-minimal.doc',
             'BFR5-Selection-v2-jef-English-minimal.html',
             'BFR5-Female-Questionnaire-v13-jef-English-minimal.html',
             'BFR5-Selection-v2-jef-Français-minimal.doc',
             'BFR5-Selection-v2-jef-English-full.doc',
             'BFR5-Selection-v2-jef-English-full.html']

        out_dir_ls_input = sorted(expected_output_unsorted)
        expected_output = sorted(out_dir_ls_input_unsorted)

        self.assertEqual(expected_output, out_dir_ls_input)
        # self.assertEqual(str(expected_output), str(out_dir_ls_input))


class MultipleFieldLanguageDelimiterSupport(PppTest):
    """Support for both : and :: to be used as delimiter betw field & lang.

        This test checks conversion for the ':' delimiter, specifically, which
        is occasionally used in some XlsForm specs, such as SurveyCTO. The '::'
        delimiter case is already covered by other tests since they use
        mostly vanilla ODK XlsForms using the '::' delimiter.
    """

    def test_convert(self):
        self.standard_conversion_test()


class SkipPatternColRelevantOrRelevance(PppTest):
    """Allow [relevant || relevance]"""

    def test_convert(self):
        self.standard_conversion_test()


class ChoiceColNameOrValue(PppTest):
    """Allow choices worksheet col [name || value]"""

    def test_convert(self):
        self.standard_conversion_test()


class SpacesInFileName(PppTest):
    """Spaces in file name: Replace or allow"""

    def test_convert(self):
        self.standard_conversion_test()


class SurveyCtoSupport(PppTest):
    """Check for successful conversion of an actual SurveyCtoFile"""

    def test_convert(self):
        self.standard_conversion_test()


class IdStringSupport(PppTest):
    """Creates a form when the form ID is called "id_string" or "form_id"."""

    def test_convert(self):
        self.standard_conversion_test()


if __name__ == '__main__':
    PARAMS = get_args()
    TEST_SUITE = get_test_suite(TEST_PACKAGES)
    unittest.TextTestRunner(verbosity=1).run(TEST_SUITE)

    if not PARAMS.doctests_only:
        unittest.main()



"""Module for the OdkCalculate class."""


class OdkCalculate:
    """Class to represent a single ODK calculate from an XLSForm."""

    def __init__(self, row):
        """Initialize the XLSForm calculate, a single row."""
        self.row = row

    def to_html(self, *args, **kwargs):
        """Convert to html."""
        return ""

    def to_text(self, *args, **kwargs):
        """Convert to text."""
        return ""

    def __repr__(self):
        """Return a representation of this calculate."""
        return '<OdkCalculate {}>'.format(self.row['name'])



"""Reusable function and class abstractions."""


def chain(_input, funcs):
    """Execute recursive function chain on input and return it.

    Side Effects: Mutates input, funcs.

    Args:
        _input: Input of any data type to be passed into functions.
        funcs: Ordered list of funcs to be applied to input.

    Returns:
        Recusive call if any functions remaining, else the altered input.
    """
    return chain(funcs.pop(0)(_input), funcs) if funcs else _input


def immutable_chain(_input, funcs):
    """Execute recursive function chain on input and return it.

    TODO: Import copy.copy only on first iteration by using an iterator int.

    Args:
        _input: Input of any data type to be passed into functions.
        funcs: Ordered list of funcs to be applied to input.

    Returns:
        Recusive call if any functions remaining, else the altered input.
    """
    from copy import copy
    new_functions = copy(funcs)
    new_input = new_functions.pop()(copy(_input)) if funcs else None
    return chain(new_input, new_functions) if funcs else copy(_input)



"""Setup for Pypi"""
import os
from setuptools import setup, find_packages


packages = find_packages(exclude=['test'])


def get_pkg_data(pkg_name, data_dirs, extensions):
    """Recusrively get package data.

    This should be called from within setup.py's setuptools.setup() function,
    e.g. package_data={'my_package': get_pkg_data(pkg_name='my_pckage', ...)}

    Args:
        pkg_name (str): Name of package containing package data.
        data_dirs (list):
        extensions (list):

    Returns:
        list: List of glob strings where each string represents a terminal
        branch node (i.e. folder) and a glob of all files ending in designated
        extension(s).

    Examples:
        Args: get_pkg_data(pkg_name='my_package',
        src_dirs=['templates'], extensions=['html'])

        Returns: [
        'templates/*.html',
        'templates/content/*.html',
        'templates/content/prompt/*.html',
        'templates/content/prompt/inputs/*.html']
    """
    pkg_root = \
        os.path.dirname(os.path.realpath(__file__)) + '/{}/'.format(pkg_name)
    pkg_data = []
    for _dir in data_dirs:
        for ext in extensions:
            for i, j, y in os.walk(pkg_root + _dir):
                pkg_data.append(i + '/*.' + ext)
    return pkg_data


setup(
    name='odk_ppp',  # The name 'ppp' was already taken.
    version='1.2.3',
    author='Joe Flack',
    author_email='jflack@jhu.edu',
    url='http://www.pma2020.org',
    packages=packages,
    package_dir={
        'ppp': 'ppp'
    },
    package_data={
        'ppp': get_pkg_data(pkg_name='ppp',
                            data_dirs=['templates'],
                            extensions=['html', 'css'])
    },
    license='LICENSE.txt',
    description='Converts XlsForm Excel files (ODK, SurveyCTO, etc) into '
                'readable formats.',
    long_description=open('README.md').read(),
    install_requires=[
        'xlrd>=0.9.3',
        'Jinja2>=2.9.6'
        'pmix==0.2.2'
    ]
)



"""Module for the OdkPrompt class."""
from sys import stderr
import re
import textwrap

#from ppp.config import TEMPLATE_ENV
from ppp.config import get_template_env
from ppp.definitions.constants import MEDIA_FIELDS, TRUNCATABLE_FIELDS, \
    LANGUAGE_DEPENDENT_FIELDS, LANGUAGE_DEPENDENT_FIELDS_NONMEDIA_FIELDS, \
    PRESETS, IGNORE_RELEVANT_TOKEN, RELEVANCE_FIELD_TOKENS, \
    PPP_REPLACEMENTS_FIELDS
from ppp.definitions.error import OdkException, OdkChoicesError

TEMPLATE_ENV = None

def set_template_env(template):
  global TEMPLATE_ENV
  TEMPLATE_ENV = get_template_env(template)

class OdkPrompt:
    """Class to represent a single ODK prompt from an XLSForm.

    This is described in a single row of an XLSForm.

    Attributes:
        row (dict): A dictionary representation of prompt.
        choices (OdkChoices): Answer choices, if applicable.
        odktype (str): The value corresponding to the prompts ODK type.
        is_section_header (bool): Designates whether or not the prompt is a
            section header.

    Class Attributes:
        select_types (tuple): Prompt types which can accept data and include a
            list of choices.
        response_types (tuple): Prompt types which can accept data and do not
            include a list of choices.
        non_response_types (tuple): Prompt types which do not accept data.

    """

    select_types = (
        'select_one',
        'select_multiple'
    )
    response_types = (
        'integer',
        'decimal',
        'geopoint',
        'barcode',
        'image',
        'text',
        'date',
        'dateTime'
    )
    non_response_types = (
        'note',
    )

    def __init__(self, row, choices=None):
        """Initialize the XLSForm prompt (a single row of a specific type).

        Args:
            row (dict): XLSForm headers as keys, row entries as values. It is
                guaranteed to have the "simple_type" key with a value from the
                class member variables `select_types`, `response_types`, or
                `non_response_types`.
            choices (OdkChoices): Answer choices, if applicable.
        """
        self.row = row
        self.choices = choices
        self.odktype = self.row['simple_type']
        self.is_section_header = True if self.row['name'].startswith('sect_') \
            else False
        if self.odktype in OdkPrompt.select_types and self.choices is None:
            msg = 'No choices found for prompt \'{}\' of type \'{}\'.'\
                .format(self.row['name'], self.odktype)
            raise OdkChoicesError(msg)

    def __repr__(self):
        """Print representation of instance."""
        return "<OdkPrompt {}>".format(self.row['name'])

    @staticmethod
    def truncate_text(text):
        """Truncate text and add an ellipsis when text is too long.

        Args:
            text (str): The text.

        Returns:
            str: Truncated text.
        """
        if len(text) > 100:
            text = text[0:98] + ' …'
        return text

    @staticmethod
    def reformat_double_line_breaks(row):
        """Convert labels and hints from strings to lists.

        This conversion process allows for line breaks to be rendered properly
        in html.

        Args:
            row (dict): The dictionary representation of prompt.

        Returns:
            dict: Reformatted representation.
        """
        for k, v in row.items():
            if True in [k.startswith(x) for x in
                        LANGUAGE_DEPENDENT_FIELDS_NONMEDIA_FIELDS]:
                if v:
                    row[k] = v.split('\n\n')
        return row

    @staticmethod
    def reformat_default_lang_vars(row, lang):
        """Reformat default language variables.

        Reformat '::/:<language_name>' style variable names to remove the
        '::/:<language_name>', leaving just the variable.

        Args:
            row (dict): The dictionary representation of prompt.
            lang (str): The language.

        Returns:
            dict: Reformatted representation.
        """
        new_row = row.copy()
        for field in LANGUAGE_DEPENDENT_FIELDS:
            if field + '::' + lang in new_row:
                new_row[field] = new_row[field + '::' + lang]
            elif field + ':' + lang in new_row:
                new_row[field] = new_row[field + ':' + lang]
        return new_row

    # pylint: disable=too-many-branches
    @staticmethod
    def create_additional_media_fields(row, prefix):
        """Create additional media fields.

        Create 'media' field for populating list of all media for prompt.
        Create individual, language-agnostic named media fields for each type
        of media present in prompt, populated with value from corresponding
        media field of default language.

        Args:
            row (dict): The dictionary representation of prompt.
            prefix (str): Prefix for media fields allowed by ODK; 'media::/:'.

        Returns:
            dict: Reformatted representation.
        """
        fields_to_add = []
        new_row = row.copy()
        for key, _ in new_row.items():
            for field in MEDIA_FIELDS:
                if key.startswith(field):
                    if field not in row:
                        fields_to_add.append(field)
                    if field.startswith(prefix):
                        non_prefixed_mf = field.replace(prefix, '')
                        if non_prefixed_mf not in row:
                            fields_to_add.append(non_prefixed_mf)

        for field in fields_to_add:
            row[field] = ''
        if fields_to_add:
            row['media'] = []
        return row

    @staticmethod
    def set_grouped_media_field(row):
        """Populate media field with all media for prompt.

        Args:
            row (dict): The dictionary representation of prompt.

        Returns:
            dict: Reformatted representation of prompt.
        """
        new_row = row.copy()
        for key, val in new_row.items():
            for field in MEDIA_FIELDS:
                if val and key.startswith(field) \
                        and val not in new_row['media']:
                    new_row['media'].append(val)
        return new_row

    @staticmethod
    def text_relevant():  # TODO: Create this method.
        # def text_relevant(self, lang):
        """Find the relevant text for this row."""
        pass

    def truncate_fields(self, row):
        """Call truncate_text() method for all truncatable fields in component.

        Args:
            row (dict): The dictionary representation of prompt.

        Returns:
            dict: Reformatted representation of prompt.
        """
        new_row = row.copy()
        for field in TRUNCATABLE_FIELDS:
            if field in new_row:
                new_row[field + '_original'] = new_row[field]
                new_row[field] = self.truncate_text(new_row[field])
        return new_row

    def format_media_labels(self, row):
        """Format text for all media labels to be enclosed in brackets.

        Args:
            row (dict): Dictionary representation of prompt.

        Returns:
            dict: Reformatted representation.
        """
        arbitrary_media_prefixes = ['media::', 'media:']
        for arb in arbitrary_media_prefixes:
            new_row = \
                self.create_additional_media_fields(row, arb)
            for key, val in new_row.items():
                for field in MEDIA_FIELDS:
                    if key.startswith(field) and val:
                        formatted_media_label = val
                        if val[0] != '[' and val[-1] != ']':
                            formatted_media_label = '[' + val + ']'
                        row[field] = formatted_media_label
                        row[key] = formatted_media_label
                        if field.startswith(arb):
                            non_prefixed_mf = field.replace(
                                arb, '')
                            row[non_prefixed_mf] = formatted_media_label
        return row

    @staticmethod
    def _ignore_relevant(prompt):
        """If applicable, ignores relevant, setting it to an empty string.

         In cases where a preset is used which makes use of human-readable
         relevants via the ppp_relevant::<language> field of an XlsForm, this
         function looks for any IGNORE_RELEVANT_TOKEN present in the field and,
         if present, sets it to an empty string.

        Args:
            prompt (dict): Dictionary representation of prompt.

        Returns
            dict: Reformatted representation.
        """
        for x in RELEVANCE_FIELD_TOKENS:
            if x in prompt:
                if prompt[x] == IGNORE_RELEVANT_TOKEN:
                    prompt[x] = ''
        return prompt

    @staticmethod
    def _extract_question_numbers(prompt):
        """Extracts question no. from label field and sets to question_number.

        First, it standardizes the label it's searching through to a string, as
        it is possible for OdkPrompt to have converted the label into a 'list'
        object. Then, it uses a regexp (regular expression) to look for a
        substring that matches a 'question number pattern', which is (1) a
        'question number' (can include letters and some other special
        characters), followed by (2) a period, followed by (3) 1 or more
        spaces, followed by (4) a letter. After this, it then sets the question
        number to be equal to the text matching (1) as just described.

        Args:
            prompt (dict): Dictionary representation of prompt.

        Returns
            dict: Reformatted representation.
        """
        # TODO: Won't need if better regexp. Current one matches sentences with
        # multiple spaces and include a number in them.
        arbitrary_character_threshold = 20
        warning = ''
        prompt['question_number'] = \
            prompt['question_number'] if 'question_number' in prompt else ''
        # This could be better. Doesn't get at default language.
        lang_labels = [x for x in prompt if x.startswith('label:')]
        label = prompt['label'] if 'label' in prompt \
            else prompt['label::English'] if 'label::English' in prompt \
            else prompt[lang_labels[0]]
        label_type = type(label).__name__
        if label_type == 'str':
            label = label
        elif label_type == 'list':
            label = label[0]
        else:
            msg = "Unsure how to handle label with type {} in the following " \
                  "label.\n\n{}.".format(label_type, label)
            raise OdkException(msg)

        match = \
            re.search(r'^(?=.*\d)[a-zA-Z0-9._\-](.+?)\.[ \n\t](.+?)[a-zA-Z].',
                      label[0:arbitrary_character_threshold])
        # Tries to match with space after '.'. If not, looks for w/ no space.
        if not match:
            match = re.search(r'^(?=.*\d)[a-zA-Z0-9._\-](.+?)\.[a-zA-Z].',
                      label[0:arbitrary_character_threshold])
            if match:
                warning = 'Warning: Question number {} does not have a space '\
                          'after \'.\''

        if match:
            q_number = match.group(0)  # gets the first match
            # removes . and anything to the right of it from q_number
            for i in range(len(q_number)):
                if q_number[-i] == '.':
                    q_number = q_number[0:-i]
                    break
            prompt['question_number'] = q_number
        if warning:
            print(warning.format(prompt['question_number']), file=stderr)

        return prompt

    @staticmethod
    def streamline_constraint_message(prompt):
        """Ensure constraint_message field name is consistent.

        Args:
            prompt (dict): Dictionary representation of prompt.

        Returns
            dict: Reformatted representation.
        """
        old, new = 'constraint message', 'constraint_message'
        if old in prompt:
            prompt[new] = prompt[old]
        return prompt

    @staticmethod
    def set_descriptive_metadata(prompt):
        """Set descriptive metadata.

        Args:
            prompt (dict): Dictionary representation of prompt.

        Returns
            dict: Reformatted representation.
        """
        prompt['is_section'] = False
        if prompt['simple_type'] == 'note' \
                and prompt['name'].startswith('sect_'):
            prompt['is_section'] = True
        return prompt

    @staticmethod
    def handle_preset(prompt, lang, preset):
        """Handle preset.

        Args:
            prompt (dict): Dictionary representation of prompt.
            lang (str): The language.
            preset (str): The preset supplied.

        Returns
            dict: Reformatted representation.
        """
        # TODO: (jef 2017.09.24) Human readable: hint variables.
        # TODO: (jef 2017.09.24) Human readable: choice filters, calcs.
        for key in prompt:
            for exclusion in PRESETS[preset]['field_exclusions']:
                if key.startswith(exclusion):
                    prompt[key] = ''
                    continue

            for to_replace in PRESETS[preset]['field_replacements']:
                replace_withs = ['ppp_'+to_replace+'::'+lang,
                                 'ppp_'+to_replace+':'+lang]
                for replace_with in replace_withs:
                    if key == replace_with and prompt[replace_with]:
                        for x in PPP_REPLACEMENTS_FIELDS:
                            if to_replace.startswith(x):
                                prompt[to_replace] = [prompt[replace_with]]

            if 'choice names' in PRESETS[preset]['other_specific_exclusions']:
                if key == 'input_field' and prompt['simple_type'] in \
                        ('select_one', 'select_multiple'):
                    prompt['input_field'] = [
                        {'name': '', 'value': '', 'label': i['label']}
                        for i in prompt['input_field']
                    ]

        OdkPrompt._ignore_relevant(prompt)

        return prompt

    def text_field(self, field, lang):
        """Find a row value given a field and language.

        An example of field and language might be "label" and "English".

        Args:
            field (str): The field from the header row.
            lang (str): The language.

        Returns:
            str: The value found from this row.
        """
        value = None
        try:
            if lang:
                if '{}::{}'.format(field, lang) in self.row:
                    key = '{}::{}'.format(field, lang)
                else:
                    key = '{}:{}'.format(field, lang)
                value = self.row[key]
            else:
                keys = (k for k in self.row if k.startswith(field))
                first = sorted(keys)[0]
                value = self.row[first]
        except (KeyError, IndexError):
            # KeyError: self.row does not have the key '{}::/:{}'.
            # IndexError: `keys` (filtered by field) is empty list.
            pass
        return value

    def to_text_relevant(self, lang):
        """Get the relevant text for this prompt.

        Args:
            :param lang: (str) The language.

        Returns:
            str: The text representation of the relevant.
        """
        formatted_relevant = None
        relevant_text = self.text_field('relevant_text', lang)
        if relevant_text:
            formatted_relevant = '[{}]'.format(relevant_text).rjust(50)
        return formatted_relevant

    def to_text_response(self, lang, numbered=False):
        """Get the response field for this prompt.

        This is a text representation of the area of a paper questionnaire
        where the response is recorded.

        Args:
            lang (str): The language.
            numbered (bool): Should choice options be numbered?

        Returns:
            str: The text representation of the response entry field.
        """
        text_str = None
        if self.odktype == 'select_multiple':
            choices = self.choices.labels(lang=lang)
            if numbered:
                choices = ['{}. {}'.format(i+1, c) for i, c in
                           enumerate(choices)]
            text_str = '\n'.join(('_ {}'.format(i) for i in choices))
        elif self.odktype == 'select_one':
            choices = self.choices.labels(lang=lang)
            if numbered:
                choices = ['{}. {}'.format(i+1, c) for i, c in
                           enumerate(choices)]
            text_str = '\n'.join(('* {}'.format(i) for i in choices))
        elif self.odktype in OdkPrompt.response_types:
            text_str = '_'*30 + '({})'.format(self.odktype)

        # try:
        #     question_type = self.odktype in Odkprompt.response_types or \
        #                     self.odktype in Odkprompt.select_types
        #     if text_str and self.row['read_only'] and question_type:
        #         # TODO: Fix read_only lookup.
        #         text_str = '\n'.join(('[Read only]', text_str))
        # except KeyError:  # Unable to find 'read_only'
        #     pass

        if text_str:
            text_str = textwrap.indent(text_str, '  ')

        return text_str

    def to_html_input_field(self, lang):
        """Get the response field for this prompt.

        This is a representation of the area of a paper questionnaire where
        the response is recorded.

        Args:
            lang (str): The language.

        Returns:
            str or dict: The representation of the entry field.
        """
        field = None
        if self.odktype in ['select_multiple', 'select_one']:
            field = self.choices.name_labels(lang=lang)
        elif self.odktype in OdkPrompt.response_types:
            field = '_' * 30 + '({})'.format(self.odktype)
        return field

    def to_text(self, lang):
        """Get the text representation of the full prompt.

        Args:
            lang (str): The language.

        Returns:
            str: The text from all parts of the prompt.
        """
        # Note: May not need 'relevant_text'.
        # relevant_text = self.text_field('relevant_text', lang)
        label = self.text_field('label', lang)
        hint = self.text_field('hint', lang)
        # Need done: Audio, Image, Video, Relevant
        fields = (
            self.to_text_relevant(lang),
            label,
            hint,
            self.to_text_response(lang)
        )
        text = filter(None, fields)
        result = '\n\n'.join(text)
        return result

    def to_dict(self, lang, **kwargs):
        """Get the text representation of the full prompt.

        Args:
            lang (str): The language.
            **bottom_border (bool): Renders a border at bottom of prompt. This
                is necessary for section headers followed by a group.

        Returns:
            dict: The text from all parts of the prompt.
        """
        # TODO: Refactor so that the dict row is only looped through once.
        prompt = self.format_media_labels(self.row)
        prompt = self.set_grouped_media_field(prompt)
        prompt = self.set_descriptive_metadata(prompt)
        prompt = self.reformat_default_lang_vars(prompt, lang)
        prompt = self.truncate_fields(prompt)
        prompt = self.reformat_double_line_breaks(prompt)
        prompt = self.streamline_constraint_message(prompt)
        prompt = self._extract_question_numbers(prompt)

        prompt['input_field'] = self.to_html_input_field(lang)

        if self.is_section_header:
            prompt['is_section_header'] = True
        if 'bottom_border' in kwargs:
            prompt['bottom_border'] = True
        if 'preset' in kwargs:
            prompt = self.handle_preset(prompt, lang, kwargs['preset'])
        return prompt

    @staticmethod
    def html_options(**kwargs):
        """HTML options.

        Args:
            **kwargs (dict): Keyword arguments.

        Returns:
            dict: Modified settings based on keyword arguments.
        """
        if 'preset' not in kwargs:
            return kwargs
        for k, v in PRESETS[kwargs['preset']]['render_settings']['html']\
                .items():
            kwargs[k] = v
        return kwargs

    def to_html(self, lang, highlighting, **kwargs):
        """Convert to html.

        Args:
            lang (str): The language.
            highlighting (bool): For color highlighting of various components
                of html template.
            **kwargs: Arbitrary keyword arguments delegated fully to to_dict().

        Returns:
            str: A rendered html template.
        """
        settings = self.html_options(**kwargs)
        # pylint: disable=no-member
        return TEMPLATE_ENV.get_template('content/content-tr-base.html')\
            .render(question=self.to_dict(lang=lang, **settings),
                    highlighting=highlighting, **settings)



"""Module for the OdkTable class."""
#from ppp.config import TEMPLATE_ENV
from ppp.config import get_template_env
from ppp.definitions.utils import exclusion
# from ppp.definitions.error import OdkformError

TEMPLATE_ENV = None

def set_template_env(template):
  global TEMPLATE_ENV
  TEMPLATE_ENV = get_template_env(template)

class OdkTable:
    """Class to represent a single ODK table from an XLSForm.

    Attributes:
        data (list): List of 1 OdkPrompt header and 1+ OdkPrompt rows.
        header (OdkPrompt): OdkPrompt representing table header.
        contents (list): List of OdkPrompts consisting of table rows.
        in_repeat (bool): Is this table part of a repeat group?
    """

    def __init__(self):
        """Initialize table object with empty initial values."""
        self.data = []
        self.header = None
        self.contents = None
        self.in_repeat = False

    def __repr__(self):
        """Print representation of instance."""
        return '<OdkTable w/ Header \'{}\': {}>'\
            .format(self.data[0].row['name'], self.data)

    def add(self, odkprompt):
        """Add a row of data from XLSForm.

        Args:
            odkprompt (OdkPrompt): ODK table row.
        """
        self.data.append(odkprompt)

    @staticmethod
    def format_row(prompt, lang, **kwargs):
        """Format rows row based on HTML options determined by kwargs.

        Args:
            prompt (OdkPrompt): The row.
            lang (str): The language.
            **kwargs: Keyword arguments.

        Returns:
            dict: Reformatted row.
        """
        settings = prompt.html_options(**kwargs)
        table_row = prompt.to_dict(lang=lang, **settings)
        return table_row

    def set_header_and_contents(self, lang, **kwargs):
        """Set header and contents of table.

        Args:
            lang (str): The language.
            **kwargs: Keyword arguments
        """
        for i in self.data:
            i.row['in_group'] = True
            i.row = self.format_row(prompt=i, lang=lang, **kwargs)
        self.header = self.data[0]
        self.contents = self.data[1:]

        # - De-list labels
        for con in self.contents:
            con.row['label'] = con.row['label'][0]

    # Temporary noinspection until method is added.
    # noinspection PyUnusedLocal
    @staticmethod
    def to_text():
        """Get the text representation of the table."""
        # def to_text(self, lang):
        # """Get the text representation of the table.
        #
        # Args:
        #     lang (str): The language.
        # Returns:
        #     str: The text for this table.
        #
        # """
        # choices = pmix.utils.d(self.choices, lang)
        #
        # choice_width = max(len(c) for c in self.choices)
        # prompt_width = max(len(p) for p in self.prompts)
        #
        # choice_format = '{:>{}}'.format(choice_width)
        # choice_labels = (choice_format.format(c) for c in self.choices)
        # choice_row = ' '.join((' ' * prompt_width, choice_labels))
        #
        # prompt_format = '{:<{}}'.format(prompt_width)
        # prompt_labels = (prompt_format.format(p) for p in self.prompts)
        #
        # option_labels = []
        # for prompt in self.prompts:
        #     if prompt.odktype == 'select_one':
        #         char = '*'
        #     elif prompt.odktype == 'select_multiple':
        #         char = '_'
        #     else:
        #         m = 'Unexpected type in ODK table: {}'.format(prompt.odktype)
        #         raise OdkformError(m)
        #     these_choices = (choice_format.format(char) for _
        # in self.choices)
        #     these_labels = ' '.join(these_choices)
        #     option_labels.append(these_labels)
        #
        # full_prompts = (' '.join(i) for i in zip(prompt_labels,
        # option_labels))
        # body = '\n'.join(full_prompts)
        # result = '\n'.join((choice_row, body))
        result = 'ODK TABLE TEXT'  # Placeholder
        return result

    def to_html(self, lang, highlighting, **kwargs):
        """Convert to html.

        Args:
            lang (place): The language.
            highlighting (bool): Displays highlighted sub-sections if True.
            **kwargs: Keyword arguments.

        Returns:
            str: A rendered html template.
        """
        # - Render header
        self.set_header_and_contents(lang, **kwargs)
        table = list()
        table.append(self.header.row)

        # - Render body
        for i in self.contents:
            if exclusion(item=i, settings=kwargs):
                continue

            table.append(i.row)

        # pylint: disable=no-member
        return TEMPLATE_ENV.get_template('content/table/table.html')\
            .render(table=table, lang=lang, highlighting=highlighting,
                    **kwargs)
